/* searchdialog.h

  Copyright (c) 2014, Nikolaj Schlej. All rights reserved.
  This program and the accompanying materials
  are licensed and made available under the terms and conditions of the BSD License
  which accompanies this distribution.  The full text of the license may be found at
  http://opensource.org/licenses/bsd-license.php

  THE PROGRAM IS DISTRIBUTED UNDER THE BSD LICENSE ON AN "AS IS" BASIS,
  WITHOUT WARRANTIES OR REPRESENTATIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED.

  */

#ifndef SEARCHDIALOG_H
#define SEARCHDIALOG_H

#include <QDialog>

#if QT_VERSION_MAJOR >= 6
#include <QRegularExpressionValidator>
#else
#include <QRegExpValidator>
#endif

#include "ui_searchdialog.h"

class SearchDialog : public QDialog
{
    Q_OBJECT

public:
    SearchDialog(QWidget *parent = 0);
    ~SearchDialog();
    Ui::SearchDialog* ui;

private slots:
    void setEditFocus(int index);

private:
#if QT_VERSION_MAJOR >= 6
    QRegularExpressionValidator hexValidator;
    QRegularExpressionValidator guidValidator;
#else
    QRegExpValidator hexValidator;
    QRegExpValidator guidValidator;
#endif
};

#endif // SEARCHDIALOG_H
